/*
copyright (c) 2008-2013 Kazuki Iwamoto http://www.maid.org/ iwm@maid.org

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#include <ctype.h>
#include <stdio.h>
#include <tchar.h>
#include <windows.h>
#include <shlwapi.h>


#define EXTENSION _T(".iwm-escape")


LPVOID
MemoryAlloc (SIZE_T nSize)
{
  return nSize > 0 ? HeapAlloc (GetProcessHeap (), 0, nSize) : NULL;
}


VOID
MemoryFree (LPVOID lpMemory)
{
  if (lpMemory)
    HeapFree (GetProcessHeap (), 0, lpMemory);
}


LPSTR
Load (LPCTSTR lpszFile)
{
  HANDLE hFile;
  LPSTR lpszText = NULL;

  hFile = CreateFile (lpszFile, GENERIC_READ, FILE_SHARE_READ,
                            NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hFile != INVALID_HANDLE_VALUE)
    {
      DWORD dwSize;

      dwSize = GetFileSize (hFile, NULL);
      if (dwSize != (DWORD)-1)
        {
          DWORD dwRead;

          lpszText = MemoryAlloc (dwSize + 1);
          if (ReadFile (hFile, lpszText, dwSize, &dwRead, NULL)
                                                        && dwSize == dwRead)
            {
              lpszText[dwSize] = '\0';
            }
          else
            {
              MemoryFree (lpszText);
              lpszText = NULL;
            }
        }
      CloseHandle (hFile);
    }
  return lpszText;
}


BOOL
Save (LPCTSTR lpszFile,
      LPCSTR  lpszText)
{
  BOOL fResult;
  HANDLE hFile;

  hFile = CreateFile (lpszFile, GENERIC_WRITE, FILE_SHARE_READ,
                            NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
  fResult = hFile != INVALID_HANDLE_VALUE;
  if (fResult)
    {
      DWORD dwSize, dwWrite;

      dwSize = lstrlenA (lpszText);
      if (!WriteFile (hFile, lpszText, dwSize, &dwWrite, NULL)
                                                        || dwSize != dwWrite)
        fResult = FALSE;
      CloseHandle (hFile);
    }
  return fResult;
}


LPSTR
Utf8ToAscii (LPCSTR lpszUtf8,
             SIZE_T nSize)
{
  int nWide;
  LPSTR lpszAscii = NULL;

  nWide = MultiByteToWideChar (CP_UTF8, 0, lpszUtf8, nSize, NULL, 0);
  if (nWide > 0)
    {
      LPWSTR lpszWide;

      lpszWide = MemoryAlloc (nWide * sizeof (WCHAR));
      if (MultiByteToWideChar (CP_UTF8, 0,
                                        lpszUtf8, nSize, lpszWide, nWide) > 0)
        {
          int nAscii;

          nAscii = WideCharToMultiByte (20127, 0,
                                        lpszWide, nWide, NULL, 0, NULL, NULL);
          if (nAscii > 0)
            {
              lpszAscii = MemoryAlloc (nAscii + 1);
              nAscii = WideCharToMultiByte (20127, 0,
                            lpszWide, nWide, lpszAscii, nAscii, NULL, NULL);
              if (nAscii > 0)
                {
                  lpszAscii[nAscii] = '\0';
                }
              else
                {
                  MemoryFree (lpszAscii);
                  lpszAscii = NULL;
                }
            }
        }
      MemoryFree (lpszWide);
    }
  return lpszAscii;
}


BOOL
Replace (LPCTSTR lpszFile)
{
  BOOL fExist, fResult = TRUE;
  TCHAR szFile[MAX_PATH];
  LPSTR lpszText;

  lstrcpy (szFile, lpszFile);
  lstrcat (szFile, EXTENSION);
  fExist = PathFileExists (szFile);
  lpszText = Load (fExist ? szFile : lpszFile);
  if (lpszText)
    {
      int i = 0, j = 0;
      BOOL fQuote = FALSE;
      LPSTR lpszEscape, p, q;

      p = lpszText;
      q = lpszEscape = MemoryAlloc (lstrlenA (lpszText) * 4 + 1);
      while (*p != '\0')
        {
          if (fQuote)
            {
              if (*p == '\"')
                fQuote = FALSE;
              if (isspace (*p) || isprint (*p))
                {
                  if (*p == '\\' && (*(p + 1) == '\"' || *(p + 1) == '\\'))
                    *q++ = *p++;
                  *q++ = *p++;
                }
              else
                {
                  *q++ = '\\';
                  switch (*p)
                    {
                      case '\n': *q++ = 'n'; break;
                      case '\t': *q++ = 't'; break;
                      case '\v': *q++ = 'v'; break;
                      case '\b': *q++ = 'b'; break;
                      case '\r': *q++ = 'r'; break;
                      case '\f': *q++ = 'f'; break;
                      case '\a': *q++ = 'a'; break;
                      default:
                        sprintf (q, "%03o", (unsigned char)*p);
                        q += 3;
                    }
                  p++;
                }
            }
          else
            {
              LPSTR r;

              for (r = p; *r != '\0' && *r != '\"'; r++);
              if (*r == '\"')
                {
                  r++;
                  fQuote = TRUE;
                }
              if (p < r)
                {
                  LPSTR lpszAscii;

                  lpszAscii = Utf8ToAscii (p, r - p);
                  lstrcpyA (q, lpszAscii);
                  q += lstrlenA (lpszAscii);
                  MemoryFree (lpszAscii);
                  p = r;
                }
            }
        }
      *q = '\0';
      if (lstrcmpA (lpszText, lpszEscape) != 0)
        {
          if ((fExist || MoveFile (lpszFile, szFile))
                                                && Save (lpszFile, lpszEscape))
            _ftprintf (stderr, _T("%s\n"), lpszFile);
          else
            fResult = FALSE;
        }
      MemoryFree (lpszEscape);
      MemoryFree (lpszText);
    }
  return fResult;
}


BOOL
Parse (LPCTSTR lpszPath,
       BOOL    fClean)
{
  BOOL fResult = TRUE;
  DWORD dwAttribute;

  dwAttribute = GetFileAttributes (lpszPath);
  if (dwAttribute != (DWORD)-1)
    {
      if (dwAttribute & FILE_ATTRIBUTE_DIRECTORY)
        {
          HANDLE hFind;
          WIN32_FIND_DATA rcFindFileData;
          TCHAR szPath[MAX_PATH];

          lstrcpy (szPath, lpszPath);
          PathAppend (szPath, _T("*"));
          hFind = FindFirstFile (szPath, &rcFindFileData);
          if (hFind != INVALID_HANDLE_VALUE)
            {
              do
                {
                  if (lstrcmp (rcFindFileData.cFileName, _T(".")) != 0
                   && lstrcmp (rcFindFileData.cFileName, _T("..")) != 0)
                    {
                      lstrcpy (szPath, lpszPath);
                      PathAppend (szPath, rcFindFileData.cFileName);
                      fResult = Parse (szPath, fClean);
                    }
                }
              while (fResult && FindNextFile (hFind, &rcFindFileData));
              if (!FindClose (hFind))
                fResult = FALSE;
            }
        }
      else
        {
          LPTSTR lpszExtension;

          lpszExtension = PathFindExtension (lpszPath);
          if (lpszExtension)
            {
              if (fClean)
                {
                  if (lstrcmpi (lpszExtension, EXTENSION) == 0)
                    {
                      TCHAR szPath[MAX_PATH];

                      lstrcpy (szPath, lpszPath);
                      lpszExtension = PathFindExtension (szPath);
                      *lpszExtension = '\0';
                      if (DeleteFile (szPath) && MoveFile (lpszPath, szPath))
                        _ftprintf (stderr, _T("%s\n"), lpszPath);
                      else
                        fResult = FALSE;
                    }
                }
              else
                {
                  if (lstrcmpi (lpszExtension, _T(".c")) == 0
                   || lstrcmpi (lpszExtension, _T(".h")) == 0)
                    fResult = Replace (lpszPath);
                }
              if (!fResult)
                _ftprintf (stderr, _T("Error:%s\n"), lpszPath);
            }
        }
    }
  return fResult;
}


int
_tmain (int    argc,
        TCHAR *argv[])
{
  int i;
  BOOL fClean = FALSE;

  for (i = 1; i < argc; i++)
    if (lstrcmpi (argv[i], _T("--clean")) == 0)
      fClean = TRUE;
  for (i = 1; i < argc; i++)
    if (*argv[i] != '-' && !Parse (argv[i], fClean))
      return -1;
  return 0;
}
